@@ -1,5 +1,6 @@
 """skimpy provides summary statistics about variables in pandas data frames."""
 from collections import defaultdict
+from itertools import chain
 
 import numpy as np
 import pandas as pd
@@ -15,7 +16,7 @@ from typeguard import typechecked
 
 console = Console()
 QUANTILES = [0, 0.25, 0.75, 1]
-HIST_BINS = 5
+HIST_BINS = 6
 UNICODE_HIST = {
     0: " ",
     1 / 8: "▁",
@@ -29,6 +30,60 @@ UNICODE_HIST = {
 }
 
 
+@typechecked
+def infer_datatypes(df: pd.DataFrame) -> pd.DataFrame:
+    """Infers the, and applies new, datatypes of dataframe columns.
+
+    :param df: input dataframe of ambiguous col type
+    :type df: pd.DataFrame
+    :return: dataframe with column datatypes set to best of knowledge
+    :rtype: pd.DataFrame
+    """
+    df_types = (
+        pd.DataFrame(df.apply(pd.api.types.infer_dtype, axis=0))
+        .reset_index()
+        .rename(columns={"index": "column", 0: "type"})
+    )
+    loop_types = df_types.values.tolist()
+    for col in loop_types:
+        if col[1] == "mixed":
+            pass
+        else:
+            if col[1] == "decimal":
+                data_type = "float64"
+            elif col[1] == "string":
+                data_type = "string"
+            elif col[1] == "integer":
+                data_type = "int"
+            elif col[1] == "floating":
+                data_type = "float64"
+            elif col[1] == "date":
+                data_type = "datetime64"
+            elif col[1] == "categorical":
+                data_type = "category"
+            elif col[1] == "boolean":
+                data_type = "bool"
+            else:
+                data_type = col[1]
+            df[col[0]] = df[col[0]].astype(data_type)
+    return df
+
+
+@typechecked
+def round_dataframe(df: pd.DataFrame) -> pd.DataFrame:
+    """Rounds dataframe to 2 s.f.
+
+    Args:
+        df (pd.DataFrame): Input dataframe
+
+    Returns:
+        pd.DataFrame: Dataframe with numbers rounded to 2 s.f.
+    """
+    for col in df.select_dtypes("number"):
+        df[col] = df[col].apply(lambda x: float(f'{float(f"{x:.2g}"):g}'))
+    return df
+
+
 @typechecked
 def dataframe_to_rich_table(
     table_name: str,
@@ -84,11 +139,13 @@ def dataframe_to_rich_table(
             ],
         )
     )
-    type_by_col = defaultdict(list)
+    type_by_col_default = defaultdict(list)
     for k, seq in cols_to_cat_map.items():
         for letter in seq:
-            type_by_col[letter].append(k)
-    type_by_col = dict(zip(type_by_col.keys(), [x[0] for x in type_by_col.values()]))
+            type_by_col_default[letter].append(k)
+    type_by_col = dict(
+        zip(type_by_col_default.keys(), [x[0] for x in type_by_col_default.values()])
+    )
     type_to_colour = dict(
         zip(type_by_col.keys(), [datatype_colours[x] for x in type_by_col.values()])
     )
@@ -97,6 +154,22 @@ def dataframe_to_rich_table(
         zip(range(len(columns)), [type_to_colour[x] for x in columns])
     )
     rows = df.values
+    # find any datetimes
+    if ("first" or "last") in df.columns:
+        timestamp_positions = [
+            [
+                [idx, i]
+                for i, j in enumerate(item)
+                if type(j) == pd._libs.tslibs.timestamps.Timestamp
+            ]
+            for idx, item in enumerate(rows)
+        ]
+        timestamp_pos_list = list(chain.from_iterable(timestamp_positions))
+        timestamp_pos_tuples = [tuple(entry) for entry in timestamp_pos_list]
+        for entry in timestamp_pos_tuples:
+            hour, min, sec = rows[entry].hour, rows[entry].minute, rows[entry].second
+            if hour == min == sec == 0:
+                rows[entry] = rows[entry].strftime("%Y-%m-%d")
     for col in columns:
         table.add_column(str(col), overflow="fold")
     for row in rows:
@@ -254,12 +327,23 @@ def string_variable_summary_table(xf: pd.DataFrame) -> pd.DataFrame:
             dict(
                 zip(
                     xf.columns,
-                    [xf[col].str.len().sum() / len(xf) for col in xf.columns],
+                    [
+                        xf[xf.columns[0]].str.count(" ").add(1).sum() / len(xf)
+                        for col in xf.columns
+                    ],
                 )
             )
         ),
         "total words": pd.Series(
-            dict(zip(xf.columns, [xf[col].str.len().sum() for col in xf.columns]))
+            dict(
+                zip(
+                    xf.columns,
+                    [
+                        xf[xf.columns[0]].str.count(" ").add(1).sum()
+                        for col in xf.columns
+                    ],
+                )
+            )
         ),
     }
     summary_df = pd.DataFrame(data_dict)
@@ -328,6 +412,9 @@ def skim(
     else:
         name = "dataframe"
 
+    # Perform inference of datatypes
+    # df = infer_datatypes(df)
+
     # Data summary
     tab_1_data = {"Number of rows": df.shape[0], "Number of columns": df.shape[1]}
     dat_sum_table = Table(
@@ -371,7 +458,9 @@ def skim(
         if not xf.empty:
             sum_df = summary_func(xf)
             list_of_tabs.append(
-                dataframe_to_rich_table(col_type, sum_df.round(2), **colour_kwargs)
+                dataframe_to_rich_table(
+                    col_type, round_dataframe(sum_df), **colour_kwargs
+                )
             )
     # Put all of the info together
     grid = Table.grid(expand=True)
@@ -382,6 +471,8 @@ def skim(
     grid.add_column(justify="left")
     for sum_tab in list_of_tabs:
         grid.add_row(sum_tab)
+    # Weirdly, iteration over list of tabs misses last entry
+    grid.add_row(list_of_tabs[-1])
     console.print(Panel(grid, title="skimpy summary", subtitle="End"))
 
 
@@ -397,32 +488,22 @@ def generate_test_data() -> pd.DataFrame:
     """
     seed = 34729
     rng = Generator(PCG64(seed))
-    columns_df = ["sepal_length", "sepal_width", "petal_length", "petal_width", "class"]
-    df = pd.DataFrame(
-        [
-            [0, 3.5, 1.4, 0.2, "setosa"],
-            [0, 3, 1.4, 0.2, "setosa"],
-            [0, 3.2, 1.3, 0.2, "setosa"],
-            [1, 3.1, 1.5, 0.2, "setosa"],
-            [1, 3.6, 1.4, 0.2, "setosa"],
-            [2, 3.9, 1.7, 0.4, "setosa"],
-            [3, np.nan, 1.4, 0.3, "setosa"],
-            [2, 3.4, 1.5, 0.2, "virginica"],
-            [1, 2.9, 1.4, 0.2, "virginica"],
-            [1, 3.1, 1.5, 0.1, "virginica"],
-            [0, 3.7, 1.5, 0.2, "virginica"],
-            [0, 3.4, 1.6, 0.2, "virginica"],
-            [0, 3, 1.4, 0.1, "virginica"],
-        ],
-        columns=columns_df,
-    )
-    df["rand_flower"] = rng.normal(size=(len(df)), scale=0.1, loc=0)
+    len_df = 1000
+    df = pd.DataFrame()
+    df["length"] = rng.beta(0.5, 0.5, size=len_df)
+    df["width"] = rng.gamma(1, 2, size=len_df)
+    df["depth"] = rng.poisson(10, size=len_df)
+    df["rnd"] = rng.normal(size=len_df, scale=1, loc=0)
+    nan_places = rng.choice(range(len_df), size=125)
+    df.loc[nan_places, "rnd"] = np.nan
+    df["class"] = rng.choice(["setosa", "virtginica"], size=len_df)
+    df["class"] = df["class"].astype("category")
     second_cat_var_entries = ["UK", "Mexico", "USA", "India"]
-    df["location"] = rng.choice(second_cat_var_entries, len(df))
+    prob = [0.6, 0.2, 0.1, 0.1]
+    df["location"] = rng.choice(second_cat_var_entries, len_df, p=prob)
     df["location"] = df["location"].astype("category")
     df.loc[3, "location"] = np.nan
-    df["class"] = df["class"].astype("category")
-    df["booly_col"] = rng.choice([True, False], size=(len(df)))
+    df["booly_col"] = rng.choice([True, False], size=len_df)
     df["booly_col"] = df["booly_col"].astype(bool)
     # string column
     string_options = [
@@ -430,13 +511,13 @@ def generate_test_data() -> pd.DataFrame:
         "What weather!",
         "Indeed, it was the most outrageously pompous cat I have ever seen.",
     ]
-    df["text"] = rng.choice(string_options, len(df))
-    df.loc[[3, 5, 8, 9], "text"] = None
+    df["text"] = rng.choice(string_options, len_df)
+    df.loc[[3, 5, 8, 9, 14, 22], "text"] = None
     df["text"] = df["text"].astype("string")
     # add a datetime column
-    df["date"] = pd.date_range("2018-01-01", periods=len(df), freq="M")
+    df["date"] = pd.date_range("2018-01-01", periods=len_df, freq="M")
     df["date_no_freq"] = rng.choice(
-        (pd.to_datetime(pd.Series(["01/01/2022", "03/04/2023", "01/05/1992"]))), len(df)
+        (pd.to_datetime(pd.Series(["01/01/2022", "03/04/2023", "01/05/1992"]))), len_df
     )
-    df.loc[3, "date_no_freq"] = pd.NaT
+    df.loc[[3, 12, 0], "date_no_freq"] = pd.NaT
     return df
